==> glexec-0.8.10.tar.gz: glexec-0.8.10/src/glexec_account.c <==
/**
 * Copyright (c) Members of the EGEE Collaboration. 2006-2010.
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Authors:
 * 2009-
 *      Oscar Koeroo <okoeroo@nikhef.nl>
 *      Mischa Sall\'e <msalle@nikhef.nl>
 *      NIKHEF Amsterdam, the Netherlands
 *
 *  2006-2009
 *      Gerben Venekamp <venekamp@nikhef.nl>
 *      NIKHEF Amsterdam, the Netherlands
 *
 */

/*!
 *  \file glexec_account.c
 *
 *  Provides helper methods to retrieve account information using getpwnam_r
 *  etc. but using caching.
 */

/* Solaris has native non-POSIX version for getpwnam_r() and getpwuid_r() */
#ifdef sun
#define _POSIX_PTHREAD_SEMANTICS    1
#endif

#include "glexec_account.h"
#include "glexec_util.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <pwd.h>
#include <grp.h>
#include <unistd.h>
#include <string.h>

/************************************************************************/
/* TYPE DEFINITIONS							*/
/************************************************************************/

struct glexec_passwd {
  struct passwd *pw;
  char *strings_buffer;
  struct glexec_passwd *next;
};

struct glexec_group {
  struct group *gr;
  struct glexec_group *next;
};

/************************************************************************/
/* GLOBAL VARIABLES							*/
/************************************************************************/

static struct glexec_passwd *glexec_passwd_first_item = NULL;
static struct glexec_group  *glexec_group_first_item  = NULL;

/************************************************************************/
/* FUNCTION PROTOTYPES							*/
/************************************************************************/

/* Creates a new glexec password struct for given pw struct (used by
 * glexec_cache_passwd_uid) */
struct glexec_passwd *glexec_create_new_glexec_passwd(struct passwd *pw);

/* Do a lookup for uid and cache the result */
struct passwd *glexec_cache_passwd_uid(struct glexec_passwd *last_item,
					uid_t uid);

/* Do a lookup for name and cache the result */
struct passwd *glexec_cache_passwd_nam(struct glexec_passwd *last_item,
				       const char *name);

/* Cache the group struct */
struct group *glexec_cache_group(struct glexec_group *last_item,
				 const struct group *gr);

/* Copy group struct */
int glexec_copy_group(const struct group *from, struct group *to);

/************************************************************************/
/* FUNCTIONS								*/
/************************************************************************/

/*!
 *  Get the passwd structure based on the uid of a user.
 *
 *  Use this function instead of the getpwuid() function as it
 *  provides caching of information. An additional benefit of this
 *  is that once a pointer has been obtained, it remains valid no
 *  matter how many times this function is called. This in contrast
 *  of calling getpwuid() directly, where the previous structure
 *  is reused and hence overwritten by each successive call.
 *  The list is freed using glexec_free_cached_passwds()
 *  \param uid
 *  \return passwd struct belonging to uid
 */
struct passwd *glexec_getpwuid(uid_t uid)    {
    struct glexec_passwd *glexec_pw=glexec_passwd_first_item,*last_item = NULL;

    /* Search for a cached value. */
    while (glexec_pw)
    {
        if (uid == glexec_pw->pw->pw_uid)
            return glexec_pw->pw;

        last_item = glexec_pw;
        glexec_pw = glexec_pw->next;
    }

    return glexec_cache_passwd_uid(last_item, uid);
}

/*!
 *  Get the passwd structure based on the name of a user.
 *
 *  Use this function instead of the getpwnam() function as it
 *  provides caching of information. An additional benefit of this
 *  is that once a pointer has been obtained, it remains valid no
 *  matter how many times this function is called. This in contrast
 *  of calling getpwnam() directly, where the previous structure
 *  is reused and hence overwritten by each successive call.
 *  The list is freed using glexec_free_cached_passwds()
 *  \param name
 *  \return passwd struct belonging to name
 */
struct passwd *glexec_getpwnam(const char *name) {
    struct glexec_passwd *glexec_pw=glexec_passwd_first_item,*last_item = NULL;

    /* Search for a cached value. */
    while (glexec_pw)
    {
	if ((strcmp(name, glexec_pw->pw->pw_name) == 0))
	    return glexec_pw->pw;

        last_item = glexec_pw;
        glexec_pw = glexec_pw->next;
    }

    return glexec_cache_passwd_nam(last_item, name);
}

/*!
 *  Get the group structure based on the gid of a user.
 *
 *  Use this function instead of the getgruid() function as it
 *  provides caching of information. An additional benefit of this
 *  is that once a pointer has been obtained, it remains valid no
 *  matter how many times this function is called. This in contrast
 *  of calling getgruid() directly, where the previous structure
 *  is reused and hence overwritten by each successive call.
 *  The list is freed using glexec_free_cached_groups()
 *  \param gid
 *  \return group struct belonging to gid
 */
struct group *glexec_getgrgid(gid_t gid) {
  struct glexec_group *glexec_gr = glexec_group_first_item, *last_item = NULL;
  struct group *gr;

  /* Search for a cached value. */
  while (glexec_gr != NULL) {
    if (gid == glexec_gr->gr->gr_gid) {
      /* Found one.*/
      return glexec_gr->gr;
    }

    last_item = glexec_gr;
    glexec_gr = glexec_gr->next;
  }

  /* Get the structure and allocate enough memory and put in in the list
   * if all is successful, and finally return the pointer to the structure. */
  if (!(gr = getgrgid(gid))) {
    return NULL;
  }

  return glexec_cache_group(last_item, gr);
}

/*!
 *  Get the group structure based on the name of a user.
 *
 *  Use this function instead of the getgrnam() function as it
 *  provides caching of information. An additional benefit of this
 *  is that once a pointer has been obtained, it remains valid no
 *  matter how many times this function is called. This in contrast
 *  of calling getgrnam() directly, where the previous structure
 *  is reused and hence overwritten by each successive call.
 *  The list is freed using glexec_free_cached_groups()
 *  \param name
 *  \return group struct belonging to name
 */
struct group *glexec_getgrnam(const char *name) {
  struct glexec_group *glexec_gr = glexec_group_first_item, *last_item = NULL;
  struct group *gr;

  /* Search for a cached value. */
  while (glexec_gr != NULL) {
    if (strcmp(name, glexec_gr->gr->gr_name) == 0) {
      /* Found one. */
      return glexec_gr->gr;
    }

    last_item = glexec_gr;
    glexec_gr = glexec_gr->next;
  }

  /* Get the structure and allocate enough memory and put in in the list
   * if all is successful, and finally return the pointer to the structure.*/
  if (!(gr = getgrnam(name))) {
    return NULL;
  }

  return glexec_cache_group(last_item, gr);
}

/*!
 * Free cached passwords
 */
void glexec_free_cached_passwds(void)
{
  struct glexec_passwd *next, *item = glexec_passwd_first_item;

  while (item) {

    if (item->pw)             {
	free(item->pw); item->pw=NULL;
    }
    if (item->strings_buffer) {
	free(item->strings_buffer); item->strings_buffer=NULL;
    }

    next = item->next;
    free(item);
    item = next;
  }
}


/*!
 * Free cached groups
 */
void glexec_free_cached_groups(void)
{
  struct glexec_group *next, *item = glexec_group_first_item;

  while (item) {

    if (item->gr) {
      if (item->gr->gr_name)   {
	  free(item->gr->gr_name); item->gr->gr_name=NULL;
      }
      if (item->gr->gr_passwd) {
	  free(item->gr->gr_passwd); item->gr->gr_passwd=NULL;
      }


      if (item->gr->gr_mem) {
        int i = 0;

        while(item->gr->gr_mem[i]) {
          free(item->gr->gr_mem[i]); item->gr->gr_mem[i]=NULL;
          i++;
        }
        free(item->gr->gr_mem); item->gr->gr_mem=NULL;
      }

      free(item->gr); item->gr=NULL;
    }

    next = item->next;
    free(item);
    item = next;
  }
}

/************************************************************************/
/* INTERNAL FUNCTIONS							*/
/************************************************************************/

/*!
 * Creates a new glexec password struct for given pw struct (used by
 * glexec_cache_passwd_uid)
 */
struct glexec_passwd *glexec_create_new_glexec_passwd(struct passwd *pw)
{
    struct glexec_passwd *g_pw;
    struct passwd *tmp_pw;
    char *buf;
    size_t strings_size, pw_name_size, pw_passwd_size, pw_gecos_size,
	   pw_dir_size, pw_shell_size, offset;

    pw_name_size   = strlen(pw->pw_name)   + 1;
    pw_passwd_size = strlen(pw->pw_passwd) + 1;
    pw_gecos_size  = strlen(pw->pw_gecos)  + 1;
    pw_dir_size    = strlen(pw->pw_dir)    + 1;
    pw_shell_size  = strlen(pw->pw_shell)  + 1;

    strings_size = pw_name_size + pw_passwd_size + pw_gecos_size +
		   pw_dir_size + pw_shell_size;

    /* Note gmalloc never returns an error, it just exits! */
    buf = (char *)gmalloc(sizeof(char) * strings_size);
    g_pw =(struct glexec_passwd *)gmalloc(sizeof(struct glexec_passwd));
    tmp_pw = (struct passwd *)gmalloc(sizeof(struct passwd));

    /* uid & gid */
    tmp_pw->pw_uid = pw->pw_uid;
    tmp_pw->pw_gid = pw->pw_gid;

    /* name */
    offset = 0;
    strncpy(buf, pw->pw_name, pw_name_size);
    tmp_pw->pw_name = buf;

    /* password */
    offset += pw_name_size;
    strncpy(buf + offset, pw->pw_passwd, pw_passwd_size);
    tmp_pw->pw_passwd = buf + offset;

    /* gecos */
    offset += pw_passwd_size;
    strncpy(buf + offset, pw->pw_gecos, pw_gecos_size);
    tmp_pw->pw_gecos = buf + offset;

    /* dir */
    offset += pw_gecos_size;
    strncpy(buf + offset, pw->pw_dir, pw_dir_size);
    tmp_pw->pw_dir = buf + offset;

    /* shell */
    offset += pw_dir_size;
    strncpy(buf + offset, pw->pw_shell, pw_shell_size);
    tmp_pw->pw_shell = buf + offset;

    g_pw->pw             = tmp_pw;
    g_pw->strings_buffer = buf;
    g_pw->next           = NULL;

    return g_pw;
}

/*!
 *  Cache the data structure.
 *
 */
struct passwd *glexec_cache_passwd_uid(struct glexec_passwd *last_item, uid_t uid)
{
    static int max_buffer_size = 0;
    char *buf = NULL;
    struct passwd pw, *p_pw;
    struct glexec_passwd *new_item;

    /* Check to see if the max buffer size has been determined before. */
    if (!max_buffer_size) 
        max_buffer_size = sysconf(_SC_GETPW_R_SIZE_MAX);

    /* See if the buffer has been allocated before and if there is a max
     * buffer size. If so, allocate the buffer and make sure allocation has
     * not failed, in which case return a NULL pointer to signal failure. */
    if (max_buffer_size)
	/* gmalloc never returns and error, just exits. */
	buf = (char *)gmalloc(sizeof(char) * max_buffer_size);

    /* initialise pw struct: getpwuid_r UPDATES it! */
    pw.pw_name=NULL;
    pw.pw_passwd=NULL;
    pw.pw_uid=-1;
    pw.pw_gid=-1;
    pw.pw_gecos=NULL;
    pw.pw_dir=NULL;
    pw.pw_shell=NULL;

    getpwuid_r(uid, &pw, buf, max_buffer_size, &p_pw);

    if (!p_pw) {
        free(buf);
        return NULL;
    }

    if ((new_item = glexec_create_new_glexec_passwd(p_pw))) {
        if (last_item) 
            last_item->next = new_item;
        else 
            glexec_passwd_first_item = new_item;
    } else {
        free(buf);
        return NULL;
    }
    free(buf);

    return new_item->pw;
}

/*!
 *  Cache the data structure.
 *
 */
struct passwd *glexec_cache_passwd_nam(struct glexec_passwd *last_item,
				       const char *name)
{
    static int max_buffer_size = 0;
    char *buf = NULL;
    struct passwd pw, *p_pw;
    struct glexec_passwd *new_item;

    /* Check to see if the max buffer size has been determined before. */
    if (!max_buffer_size) 
        max_buffer_size = sysconf(_SC_GETPW_R_SIZE_MAX);

    /* See if the buffer has been allocated before and if there is a max
     * buffer size. If so, allocate the buffer and make sure allocation has
     * not failed, in which case return a NULL pointer to signal failure. */
    if (max_buffer_size)
	/* gmalloc never returns and error, just exits. */
        buf = (char *)gmalloc(sizeof(char) * max_buffer_size);

    /* initialise pw struct: getpwuid_r UPDATES it! */
    pw.pw_name=NULL;
    pw.pw_passwd=NULL;
    pw.pw_uid=-1;
    pw.pw_gid=-1;
    pw.pw_gecos=NULL;
    pw.pw_dir=NULL;
    pw.pw_shell=NULL;

    getpwnam_r(name, &pw, buf, max_buffer_size, &p_pw);

    if (!p_pw) {
        free(buf);
        return NULL;
    }

    if ((new_item = glexec_create_new_glexec_passwd(p_pw))) {
        if (last_item) 
            last_item->next = new_item;
        else 
            glexec_passwd_first_item = new_item;
    } else {
        free(buf);
        return NULL;
    }
    free(buf);

    return new_item->pw;
}

/*!
 *  Cache the data structure.
 */
struct group *glexec_cache_group(struct glexec_group *last_item,
				 const struct group *gr)
{
    struct glexec_group *new_item;
    struct group *tmp_gr;

    /* gmalloc never returns an error, just exits */
    tmp_gr = gmalloc(sizeof(struct group));
    new_item = gmalloc(sizeof(struct glexec_group));

    new_item->gr   = tmp_gr;
    new_item->next = NULL;

    /* Either succesful, or never returns...*/
    glexec_copy_group(gr, tmp_gr);

    if (last_item)
	last_item->next = new_item;
    else
	glexec_group_first_item = new_item;

    return tmp_gr;
}


/*!
 *  Copy a group structure.
 */
int glexec_copy_group(const struct group *from, struct group *to)
{
    char **members;
    int i, n;

    /* gstrdup never returns an error, just exits. It can return NULL, but not
     * in error, only because argument is also NULL */
    to->gr_name = gstrdup(from->gr_name);
    to->gr_passwd = gstrdup(from->gr_passwd);
    to->gr_gid = from->gr_gid;

    /* Find number of members to allocate memory */
    members = from->gr_mem;
    for (n=0; members[n] != NULL ; n++);
    
    /* gmalloc never returns an error, just exits */
    to->gr_mem = (char **)gmalloc((n+1)*sizeof(char *));

    /* copy group members, this INCLUDES the last NULL */
    /* Note that gstrdup never returns an error, just exits */
    for (i=0; i<=n; i++)
	to->gr_mem[i] = gstrdup(from->gr_mem[i]);

    return 0;
}

==> glexec_0.9.11.orig.tar.gz: glexec-0.9.11/src/glexec_account.c <==
/**
 * Copyright (c) Members of the EGEE Collaboration. 2006-2010.
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Authors:
 * 2009-
 *      Oscar Koeroo <okoeroo@nikhef.nl>
 *      Mischa Sall\'e <msalle@nikhef.nl>
 *      NIKHEF Amsterdam, the Netherlands
 *
 *  2006-2009
 *      Gerben Venekamp <venekamp@nikhef.nl>
 *      NIKHEF Amsterdam, the Netherlands
 *
 */

/*!
 *  \file glexec_account.c
 *
 *  Provides helper functions to retrieve account information using getpwnam_r
 *  etc. but using caching.
 */

/* Solaris has native non-POSIX version for getpwnam_r() and getpwuid_r() */
#ifdef __sun
# define _POSIX_PTHREAD_SEMANTICS    1
#endif

#define _XOPEN_SOURCE	500

#include "glexec_account.h"
#include "glexec_util.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <pwd.h>
#include <grp.h>
#include <unistd.h>
#include <string.h>

/************************************************************************/
/* TYPE DEFINITIONS							*/
/************************************************************************/

struct glexec_passwd {
  struct passwd *pw;
  char *strings_buffer;
  struct glexec_passwd *next;
};

struct glexec_group {
  struct group *gr;
  struct glexec_group *next;
};

/************************************************************************/
/* GLOBAL VARIABLES							*/
/************************************************************************/

static struct glexec_passwd *glexec_passwd_first_item = NULL;
static struct glexec_group  *glexec_group_first_item  = NULL;

/************************************************************************/
/* FUNCTION PROTOTYPES							*/
/************************************************************************/

/* Creates a new glexec password struct for given pw struct (used by
 * glexec_cache_passwd_uid) */
static struct glexec_passwd *glexec_create_new_glexec_passwd(struct passwd *pw);

/* Do a lookup for uid and cache the result */
static struct passwd *glexec_cache_passwd_uid(struct glexec_passwd *last_item,
                                              uid_t uid);

/* Do a lookup for name and cache the result */
static struct passwd *glexec_cache_passwd_nam(struct glexec_passwd *last_item,
                                              const char *name);

/* Cache the group struct */
static struct group *glexec_cache_group(struct glexec_group *last_item,
                                        const struct group *gr);

/* Copy group struct */
static int glexec_copy_group(const struct group *from, struct group *to);

/************************************************************************/
/* FUNCTIONS								*/
/************************************************************************/

/*!
 *  Get the passwd structure based on the uid of a user.
 *
 *  Use this function instead of the getpwuid() function as it
 *  provides caching of information. An additional benefit of this
 *  is that once a pointer has been obtained, it remains valid no
 *  matter how many times this function is called. This in contrast
 *  of calling getpwuid() directly, where the previous structure
 *  is reused and hence overwritten by each successive call.
 *  The list is freed using glexec_free_cached_passwds()
 *  \param uid
 *  \return passwd struct belonging to uid
 */
struct passwd *glexec_getpwuid(uid_t uid)    {
    struct glexec_passwd *glexec_pw=glexec_passwd_first_item,*last_item = NULL;

    /* Search for a cached value. */
    while (glexec_pw)
    {
        if (uid == glexec_pw->pw->pw_uid)
            return glexec_pw->pw;

        last_item = glexec_pw;
        glexec_pw = glexec_pw->next;
    }

    return glexec_cache_passwd_uid(last_item, uid);
}

/*!
 *  Get the passwd structure based on the name of a user.
 *
 *  Use this function instead of the getpwnam() function as it
 *  provides caching of information. An additional benefit of this
 *  is that once a pointer has been obtained, it remains valid no
 *  matter how many times this function is called. This in contrast
 *  of calling getpwnam() directly, where the previous structure
 *  is reused and hence overwritten by each successive call.
 *  The list is freed using glexec_free_cached_passwds()
 *  \param name
 *  \return passwd struct belonging to name
 */
struct passwd *glexec_getpwnam(const char *name) {
    struct glexec_passwd *glexec_pw=glexec_passwd_first_item,*last_item = NULL;

    /* Search for a cached value. */
    while (glexec_pw)
    {
	if ((strcmp(name, glexec_pw->pw->pw_name) == 0))
	    return glexec_pw->pw;

        last_item = glexec_pw;
        glexec_pw = glexec_pw->next;
    }

    return glexec_cache_passwd_nam(last_item, name);
}

/*!
 *  Get the group structure based on the gid of a user.
 *
 *  Use this function instead of the getgruid() function as it
 *  provides caching of information. An additional benefit of this
 *  is that once a pointer has been obtained, it remains valid no
 *  matter how many times this function is called. This in contrast
 *  of calling getgruid() directly, where the previous structure
 *  is reused and hence overwritten by each successive call.
 *  The list is freed using glexec_free_cached_groups()
 *  \param gid
 *  \return group struct belonging to gid
 */
struct group *glexec_getgrgid(gid_t gid) {
  struct glexec_group *glexec_gr = glexec_group_first_item, *last_item = NULL;
  struct group *gr;

  /* Search for a cached value. */
  while (glexec_gr != NULL) {
    if (gid == glexec_gr->gr->gr_gid) {
      /* Found one.*/
      return glexec_gr->gr;
    }

    last_item = glexec_gr;
    glexec_gr = glexec_gr->next;
  }

  /* Get the structure and allocate enough memory and put in in the list
   * if all is successful, and finally return the pointer to the structure. */
  if (!(gr = getgrgid(gid))) {
    return NULL;
  }

  return glexec_cache_group(last_item, gr);
}

/*!
 *  Get the group structure based on the name of a user.
 *
 *  Use this function instead of the getgrnam() function as it
 *  provides caching of information. An additional benefit of this
 *  is that once a pointer has been obtained, it remains valid no
 *  matter how many times this function is called. This in contrast
 *  of calling getgrnam() directly, where the previous structure
 *  is reused and hence overwritten by each successive call.
 *  The list is freed using glexec_free_cached_groups()
 *  \param name
 *  \return group struct belonging to name
 */
struct group *glexec_getgrnam(const char *name) {
  struct glexec_group *glexec_gr = glexec_group_first_item, *last_item = NULL;
  struct group *gr;

  /* Search for a cached value. */
  while (glexec_gr != NULL) {
    if (strcmp(name, glexec_gr->gr->gr_name) == 0) {
      /* Found one. */
      return glexec_gr->gr;
    }

    last_item = glexec_gr;
    glexec_gr = glexec_gr->next;
  }

  /* Get the structure and allocate enough memory and put in in the list
   * if all is successful, and finally return the pointer to the structure.*/
  if (!(gr = getgrnam(name))) {
    return NULL;
  }

  return glexec_cache_group(last_item, gr);
}

/*!
 * Free cached passwords
 */
void glexec_free_cached_passwds(void)
{
  struct glexec_passwd *next, *item = glexec_passwd_first_item;

  while (item) {

    if (item->pw)             {
	free(item->pw); item->pw=NULL;
    }
    if (item->strings_buffer) {
	free(item->strings_buffer); item->strings_buffer=NULL;
    }

    next = item->next;
    free(item);
    item = next;
  }
}


/*!
 * Free cached groups
 */
void glexec_free_cached_groups(void)
{
  struct glexec_group *next, *item = glexec_group_first_item;

  while (item) {

    if (item->gr) {
      if (item->gr->gr_name)   {
	  free(item->gr->gr_name); item->gr->gr_name=NULL;
      }
      if (item->gr->gr_passwd) {
	  free(item->gr->gr_passwd); item->gr->gr_passwd=NULL;
      }


      if (item->gr->gr_mem) {
        int i = 0;

        while(item->gr->gr_mem[i]) {
          free(item->gr->gr_mem[i]); item->gr->gr_mem[i]=NULL;
          i++;
        }
        free(item->gr->gr_mem); item->gr->gr_mem=NULL;
      }

      free(item->gr); item->gr=NULL;
    }

    next = item->next;
    free(item);
    item = next;
  }
}

/************************************************************************/
/* INTERNAL FUNCTIONS							*/
/************************************************************************/

/*!
 * Creates a new glexec password struct for given pw struct (used by
 * glexec_cache_passwd_uid)
 */
static struct glexec_passwd *glexec_create_new_glexec_passwd(struct passwd *pw)
{
    struct glexec_passwd *g_pw;
    struct passwd *tmp_pw;
    char *buf;
    size_t strings_size, pw_name_size, pw_passwd_size, pw_gecos_size,
	   pw_dir_size, pw_shell_size, offset;

    pw_name_size   = strlen(pw->pw_name)   + 1;
    pw_passwd_size = strlen(pw->pw_passwd) + 1;
    pw_gecos_size  = strlen(pw->pw_gecos)  + 1;
    pw_dir_size    = strlen(pw->pw_dir)    + 1;
    pw_shell_size  = strlen(pw->pw_shell)  + 1;

    strings_size = pw_name_size + pw_passwd_size + pw_gecos_size +
		   pw_dir_size + pw_shell_size;

    /* Note gmalloc never returns an error, it just exits! */
    buf = (char *)gmalloc(sizeof(char) * strings_size);
    g_pw =(struct glexec_passwd *)gmalloc(sizeof(struct glexec_passwd));
    tmp_pw = (struct passwd *)gmalloc(sizeof(struct passwd));

    /* uid & gid */
    tmp_pw->pw_uid = pw->pw_uid;
    tmp_pw->pw_gid = pw->pw_gid;

    /* name */
    offset = 0;
    strncpy(buf, pw->pw_name, pw_name_size);
    tmp_pw->pw_name = buf;

    /* password */
    offset += pw_name_size;
    strncpy(buf + offset, pw->pw_passwd, pw_passwd_size);
    tmp_pw->pw_passwd = buf + offset;

    /* gecos */
    offset += pw_passwd_size;
    strncpy(buf + offset, pw->pw_gecos, pw_gecos_size);
    tmp_pw->pw_gecos = buf + offset;

    /* dir */
    offset += pw_gecos_size;
    strncpy(buf + offset, pw->pw_dir, pw_dir_size);
    tmp_pw->pw_dir = buf + offset;

    /* shell */
    offset += pw_dir_size;
    strncpy(buf + offset, pw->pw_shell, pw_shell_size);
    tmp_pw->pw_shell = buf + offset;

    g_pw->pw             = tmp_pw;
    g_pw->strings_buffer = buf;
    g_pw->next           = NULL;

    return g_pw;
}

/*!
 *  Cache the data structure.
 *
 */
static struct passwd *glexec_cache_passwd_uid(struct glexec_passwd *last_item,
                                              uid_t uid)
{
    static long int max_buffer_size = 0;
    char *buf = NULL;
    struct passwd pw, *p_pw;
    struct glexec_passwd *new_item;

    /* Check to see if the max buffer size has been determined before */
    if (max_buffer_size==0)
	max_buffer_size = sysconf(_SC_GETPW_R_SIZE_MAX);

    /* Check to see if the max buffer size has failed (before) */
    if (max_buffer_size<=0)
	return NULL;

    /* gmalloc never returns and error, just exits. max_buffer_size is
     * guaranteed > 0 */
    buf = (char *)gmalloc(sizeof(char) * (size_t)max_buffer_size);

    /* initialise pw struct: getpwuid_r UPDATES it! */
    pw.pw_name=NULL;
    pw.pw_passwd=NULL;
    pw.pw_uid=uid; /* pw_uid is uid_t which is unsigned, so init with val uid */
    pw.pw_gid=GLEXEC_GID_UNDEFINED; /* It should be overwritten or p_pw will be
				       NULL, just use this as a relatively safe
				       default */
    pw.pw_gecos=NULL;
    pw.pw_dir=NULL;
    pw.pw_shell=NULL;

    /* max_buffer_size is guaranteed > 0, see above */
#ifdef __sun
    getpwuid_r(uid, &pw, buf, (int)max_buffer_size, &p_pw);
#else
    getpwuid_r(uid, &pw, buf, (size_t)max_buffer_size, &p_pw);
#endif

    if (!p_pw) {
        free(buf);
        return NULL;
    }

    if ((new_item = glexec_create_new_glexec_passwd(p_pw))) {
        if (last_item) 
            last_item->next = new_item;
        else 
            glexec_passwd_first_item = new_item;
    } else {
        free(buf);
        return NULL;
    }
    free(buf);

    return new_item->pw;
}

/*!
 *  Cache the data structure.
 *
 */
static struct passwd *glexec_cache_passwd_nam(struct glexec_passwd *last_item,
                                              const char *name)
{
    static long int max_buffer_size = 0;
    char *buf = NULL;
    struct passwd pw, *p_pw;
    struct glexec_passwd *new_item;

    /* Check to see if the max buffer size has been determined before */
    if (max_buffer_size==0)
	max_buffer_size = sysconf(_SC_GETPW_R_SIZE_MAX);

    /* Check to see if the max buffer size has failed (before) */
    if (max_buffer_size<=0)
	return NULL;

    /* gmalloc never returns and error, just exits. max_buffer_size is
     * guaranteed > 0 */
    buf = (char *)gmalloc(sizeof(char) * (size_t)max_buffer_size);

    /* initialise pw struct: getpwuid_r UPDATES it! */
    pw.pw_name=NULL;
    pw.pw_passwd=NULL;
    pw.pw_uid=GLEXEC_UID_UNDEFINED; /* It should be overwritten or p_pw will be
				       NULL, just use this as a relatively safe
				       default */
    pw.pw_gid=GLEXEC_GID_UNDEFINED; /* It should be overwritten or p_pw will be
				       NULL, just use this as a relatively safe
				       default */
    pw.pw_gecos=NULL;
    pw.pw_dir=NULL;
    pw.pw_shell=NULL;


    /* max_buffer_size is guaranteed > 0, see above */
#ifdef __sun
    getpwnam_r(name, &pw, buf, (int)max_buffer_size, &p_pw);
#else
    getpwnam_r(name, &pw, buf, (size_t)max_buffer_size, &p_pw);
#endif

    if (!p_pw) {
        free(buf);
        return NULL;
    }

    if ((new_item = glexec_create_new_glexec_passwd(p_pw))) {
        if (last_item) 
            last_item->next = new_item;
        else 
            glexec_passwd_first_item = new_item;
    } else {
        free(buf);
        return NULL;
    }
    free(buf);

    return new_item->pw;
}

/*!
 *  Cache the data structure.
 */
static struct group *glexec_cache_group(struct glexec_group *last_item,
                                        const struct group *gr)
{
    struct glexec_group *new_item;
    struct group *tmp_gr;

    /* gmalloc never returns an error, just exits */
    tmp_gr = gmalloc(sizeof(struct group));
    new_item = gmalloc(sizeof(struct glexec_group));

    new_item->gr   = tmp_gr;
    new_item->next = NULL;

    /* Either succesful, or never returns...*/
    glexec_copy_group(gr, tmp_gr);

    if (last_item)
	last_item->next = new_item;
    else
	glexec_group_first_item = new_item;

    return tmp_gr;
}


/*!
 *  Copy a group structure.
 */
static int glexec_copy_group(const struct group *from, struct group *to)
{
    char **members;
    size_t i,n;

    /* gstrdup never returns an error, just exits. It can return NULL, but not
     * in error, only because argument is also NULL */
    to->gr_name = gstrdup(from->gr_name);
    to->gr_passwd = gstrdup(from->gr_passwd);
    to->gr_gid = from->gr_gid;

    /* Find number of members to allocate memory */
    members = from->gr_mem;
    for (n=0; members[n] != NULL ; n++);
    
    /* gmalloc never returns an error, just exits */
    to->gr_mem = (char **)gmalloc((n+1)*sizeof(char *));

    /* copy group members, this INCLUDES the last NULL */
    /* Note that gstrdup never returns an error, just exits */
    for (i=0; i<=n; i++)
	to->gr_mem[i] = gstrdup(from->gr_mem[i]);

    return 0;
}
